###############################################################################
#   rpmorphan-tk-lib.pl
#
#    Copyright (C) 2006 by Eric Gerbier
#    Bug reports to: gerbier@users.sourceforge.net
#    $Id: rpmorphan-1.18 | rpmorphan-tk-lib.pl | Fri Dec 15 13:03:02 2017 +0100 | Eric Gerbier  $
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
###############################################################################
# this file contains all tk code for rpmorphan gui
###############################################################################

use strict;
use warnings;

use English '-no_match_vars';

# cf also test_tk_gui sub in rpmorphan.pl
use lib '.';    # for local test (MListbox)
use Tk;
use Tk::Balloon;
use Tk::Dialog;
use Tk::Frame;
use Tk::Label;
use Tk::Listbox;
use Tk::Message;
use Tk::ROText;
use Tk::Event;       # virtual event, eventAdd / eventGenerate
use Tk::MListbox;    # local for columns

#use Data::Dumper;

## no critic(ProhibitPackageVars)
use vars
  qw( $Opt_dry_run $Inloop @Liste_pac &display_total &insert_gui_orphans );
## use critic

my $W_list;

################################################################
# select all
sub tk_select_callback() {

	$W_list->selectionSet( 0, 'end' );
	return;
}
################################################################
# unselect all
sub tk_unselect_callback() {

	$W_list->selectionClear( 0, 'end' );
	return;
}
#########################################################
# change cursor to watch
sub tk_cursor2watch() {

	my $cursor = $W_list->cget('-cursor');
	$W_list->configure( -cursor => 'watch' );
	$W_list->update();

	return $cursor;
}
#########################################################
# restore cursor
sub tk_cursor2normal($) {
	my $cursor = shift @_;    # saved cursor

	# restore pointer
	$W_list->configure( -cursor => $cursor );
	$W_list->update();
	return;
}
#########################################################
# return a package name
# the format depends upon arch option
sub tk_get_name(@) {
	my $r_tab = shift @_;

	my $pac;
	my $name     = @{$r_tab}[0];
	my $opt_arch = get_option('arch');
	if ($opt_arch) {
		$pac = build_full_name( $name, @{$r_tab}[1], @{$r_tab}[2] );
	}
	else {
		$pac = $name;
	}
	return $pac;
}
#########################################################
# remove selected packages
sub tk_remove_callback($) {
	my $main = shift @_;    # parent widget

	# change pointer
	my $cursor = tk_cursor2watch();

	# get full list
	my @liste = $W_list->get( 0, 'end' );

	# get selected items
	my @sel = $W_list->curselection();

	# transform index into names and get first field (name)
	my @selnames = map { tk_get_name( $liste[$_] ) } @sel;

	my $freesize = get_package_size(@selnames);

	# confirm dialog
	my $answer = tk_dialog( $main, 'confirm',
		"confirm remove @selnames ?\nwill free $freesize" );

	if ( $answer =~ m/Yes/ ) {

		remove_packages( \@selnames, \@liste );
	}

	# restore pointer
	tk_cursor2normal($cursor);
	return;
}
#########################################################
# general text display in a new text window
# is used by all help buttons
sub tk_display_message($$@) {
	my $main    = shift @_;    # parent widget
	my $title   = shift @_;    # window title
	my @baratin = @_;          # text to display

	my $top = $main->Toplevel( -title => $title );
	$top->Button( -text => 'quit', -command => [ $top => 'destroy' ] )->pack();
	my $text = $top->Scrolled(
		'ROText',
		-scrollbars => 'e',
		-height     => 25,
		-width      => 128,
		-wrap       => 'word'
	)->pack( -side => 'left', -expand => 1, -fill => 'both' );
	$top->bind( '<Key-q>', [ $top => 'destroy' ] );

	$text->insert( 'end', @baratin );
	$text->see('1.0');
	return;
}

################################################################
# general dialog
sub tk_dialog($$$) {
	my $main    = shift @_;    # parent widget
	my $title   = shift @_;    # window title
	my $baratin = shift @_;    # text to display

	my $dialog = $main->Dialog(
		-text           => $baratin,
		-default_button => 'Yes',
		-buttons        => [qw/Yes No/]
	);
	my $answer = $dialog->Show();

	return $answer;
}
################################################################
# display help text
sub tk_help_callback($) {
	my $main = shift;    # parent widget

	my $baratin = get_help_text();

	tk_display_message( $main, 'help', $baratin );
	return;
}
################################################################
# display log file
sub tk_log_callback($) {
	my $main = shift;    # parent widget

	my @l = get_log();
	my $l = join "\n", @l;

	tk_display_message( $main, 'help', $l );
	return;
}
#########################################################
# return name of current package
sub tk_get_current_elem() {

	my $opt_arch = get_option('arch');

	my $elem;
	if ($opt_arch) {
		my $listbox_name    = $W_list->columnGet(0)->Subwidget('listbox');
		my $listbox_version = $W_list->columnGet(1)->Subwidget('listbox');
		my $listbox_arch    = $W_list->columnGet(2)->Subwidget('listbox');

		$elem = build_full_name(
			$listbox_name->get('active'),
			$listbox_version->get('active'),
			$listbox_arch->get('active')
		);
	}
	else {

		# get current position
		my $listbox = $W_list->columnGet(0)->Subwidget('listbox');
		$elem = $listbox->get('active');
	}

	#print "debug : tk_get_current_elem : $elem\n";

	return $elem;
}
#########################################################
# display package summary
sub tk_summary_callback($) {
	my $main = shift @_;    # parent widget

	my $c = tk_get_current_elem();
	if ($c) {
		tk_popup( $main, get_package_summary($c) );
	}
	else {
		print "no current element selected\n";
	}

	return;
}
################################################################
# display package info
sub tk_info_callback($) {
	my $main = shift @_;    # parent widget

	# get current position
	my $c = tk_get_current_elem();

	my @res = get_package_info($c);
	chomp @res;
	my $message = join "\n", @res;

	tk_display_message( $main, 'info', $message );
	return;
}
#########################################################
# for reload button
sub tk_load_callback($) {
	my $main = shift @_;    # parent widget

	load_data(1);

	return;
}
#########################################################
# is called when an item is selected/unselected
# and change a counter
sub tk_list_callback {
	my $widget   = shift @_;
	my $selected = shift @_;

	# get selected items
	my @sel = $W_list->curselection();

	my $nb = scalar @sel;

	$selected->delete( '1.0', 'end' );
	$selected->insert( '1.0', $nb );

	return;
}
#########################################################
# common sub for button and key
sub tk_quit($) {
	my $main = shift @_;
	$main->destroy();

	return;
}
#########################################################
sub tk_popup($$) {
	my $main    = shift @_;
	my $message = shift @_;

	my $top = $main->Toplevel();
	$top->title('popup');
	my $d = $top->Message( -text => $message );
	$d->pack();
	$d->focus();
	$top->update();

	# wait 1 second and destroy popup
	# no critic (ProhibitMagicNumbers)
	$main->after(1000);

	$top->destroy();

	return;
}
#########################################################
sub build_tk_gui($) {
	my $rh_opt = shift @_;

	my $main      = MainWindow->new( -title => build_title() );
	my $w_balloon = $main->Balloon();
	my $side      = 'top';

	# frame for buttons
	###################
	my $frame1 = $main->Frame();
	$frame1->pack( -side => 'top', -expand => 0, -fill => 'none' );

	# remove : <<remove>>
	$main->bind( '<<remove>>', [ \&tk_remove_callback, $main ] );
	my $remove_button = $frame1->Button(
		-text    => 'Remove',
		-command => sub {
			$main->eventGenerate('<<remove>>');
		},
	);
	$remove_button->pack( -side => 'left' );
	$w_balloon->attach( $remove_button, -msg => 'remove selected packages' );
	if ( is_remove_allowed($Opt_dry_run) ) {

		$main->eventAdd( '<<remove>>' => '<Key-r>' );
		$main->eventAdd( '<<remove>>' => '<Key-x>' );
	}
	else {
		$remove_button->configure( -state => 'disabled' );
	}

	# info : use virtual event <<info>>
	$main->bind( '<<info>>', [ \&tk_info_callback, $main ] );
	my $info_button = $frame1->Button(
		-text    => 'Info',
		-command => sub {
			$main->eventGenerate('<<info>>');

			#			print "pressed button info\n";
		},
	);
	$info_button->pack( -side => 'left' );
	$w_balloon->attach( $info_button, -msg => 'get info on current package' );

	# debug
	# $main->bind( '<Key-i>', sub {
	# 	$main->eventGenerate('<<info>>');
	# 	print "pressed key i\n";
	# 	});
	$main->eventAdd( '<<info>>' => '<Key-i>' );

	# select all : <<select_all>>
	$main->bind( '<<select_all>>', [ \&tk_select_callback ] );
	my $select_button = $frame1->Button(
		-text    => 'Select all',
		-command => sub {
			$main->eventGenerate('<<select_all>>');
		}
	);
	$select_button->pack( -side => 'left' );
	$w_balloon->attach( $select_button, -msg => 'select all packages' );
	$main->eventAdd( '<<select_all>>' => '<Key-s>' );

	# unselect all : <<unselect_all>>
	$main->bind( '<<unselect_all>>', [ \&tk_unselect_callback ] );
	my $unselect_button = $frame1->Button(
		-text    => 'Unselect all',
		-command => sub {
			$main->eventGenerate('<<unselect_all>>');
		}
	);
	$unselect_button->pack( -side => 'left' );
	$w_balloon->attach( $unselect_button, -msg => 'unselect all packages' );
	$main->eventAdd( '<<unselect_all>>' => '<Key-u>' );

	# log : <<log>>
	$main->bind( '<<log>>', [ \&tk_log_callback, $main ] );
	my $log_button = $frame1->Button(
		-text    => 'Log',
		-command => sub {
			$main->eventGenerate('<<log>>');
		}
	);
	$log_button->pack( -side => 'left' );
	$w_balloon->attach( $log_button, -msg => 'display log' );
	$main->eventAdd( '<<log>>' => '<Key-l>' );

	# reload button : <<reload>>
	$main->bind( '<<reload>>', [ \&tk_load_callback, $main ] );
	my $reload_button = $frame1->Button(
		-text    => 'reLoad',
		-command => sub {
			$main->eventGenerate('<<reload>>');
		}
	);
	$reload_button->pack( -side => 'left' );
	$w_balloon->attach( $reload_button, -msg => 'reload rpm database' );
	$main->eventAdd( '<<reload>>' => '<Key-L>' );

	# help : <<help>>
	$main->bind( '<<help>>', [ \&tk_help_callback, $main ] );
	my $help_button = $frame1->Button(
		-text    => 'Help',
		-command => sub {
			$main->eventGenerate('<<help>>');
		}
	);
	$help_button->pack( -side => 'left' );
	$w_balloon->attach( $help_button, -msg => 'help on rpmorphan interface' );
	$main->eventAdd( '<<help>>' => '<Key-h>' );

	# quit : <<quit>>
	$main->bind( '<<quit>>', [ \&tk_quit, $main ] );
	my $exit_button = $frame1->Button(
		-text    => 'Quit',
		-command => sub {
			$main->eventGenerate('<<quit>>');
		}
	);
	$exit_button->pack( -side => 'left' );
	$w_balloon->attach( $exit_button, -msg => 'quit the software' );
	$main->eventAdd( '<<quit>>' => '<Key-q>' );

	# frame for the list
	####################
	my $frame2 = $main->Frame();
	$frame2->pack( -side => $side, -expand => 1, -fill => 'both' );

	my @columns;
	if ( ${ $rh_opt->{'arch'} } ) {
		push @columns, [qw/-text Name -textwidth 20/];
		push @columns, [qw/-text Version -textwidth 20/];
		push @columns, [qw/-text Arch -textwidth 20/];
	}
	else {
		push @columns, [qw/-text Name -textwidth 30/];
	}
	push @columns, [qw/-text Size -textwidth 20/];

	$W_list = $frame2->Scrolled(
		'MListbox',
		-scrollbars => 'e',
		-selectmode => 'multiple',
		-columns    => \@columns,
	)->pack( -side => $side, -expand => 1, -fill => 'both' );

	# sort numerical on size column (last columns)
	$W_list->columnConfigure( $#columns,
		-comparecommand => sub { $_[0] <=> $_[1] } );

	$w_balloon->attach( $W_list, -msg => 'list window' );

	# for summary
	$main->bind( '<Key-S>', [ \&tk_summary_callback, $main ] );

	# button 3 : full info
	# debug
	# $main->bind( '<Button-3>', sub {
	# 	$main->eventGenerate('<<info>>');
	# 	print "pressed button 3\n";
	# 	} );
	$main->eventAdd( '<<info>>' => '<Button-3>' );

	# frame for status
	###################
	my $frame3 = $main->Frame();
	$frame3->pack( -side => 'bottom', -expand => 0, -fill => 'none' );

	# number of packages
	$frame3->Label( -text => 'total' )->pack( -side => 'left' );
	my $w_total = $frame3->ROText( -width => 4, -height => 1 );
	$w_total->pack( -side => 'left', -expand => 0, );
	$w_balloon->attach( $w_total, -msg => 'number of packages' );

	# number of orphans
	$frame3->Label( -text => 'orphans' )->pack( -side => 'left' );
	my $w_orphans = $frame3->ROText( -width => 4, -height => 1 );
	$w_orphans->pack( -side => 'left', -expand => 0, );
	$w_balloon->attach( $w_orphans, -msg => 'number of orphans packages' );

	# number of selected
	$frame3->Label( -text => 'selected' )->pack( -side => 'left' );
	my $w_selected = $frame3->ROText( -width => 4, -height => 1 );
	$w_selected->pack( -side => 'left', -expand => 0, );
	$w_balloon->attach( $w_selected, -msg => 'number of selected orphans' );

	# to allow to show number of selected
	my $listbox = $W_list->columnGet(0)->Subwidget('listbox');
	$listbox->bind( '<<ListboxSelect>>', [ \&tk_list_callback, $w_selected ] );

	# status
	$frame3->Label( -text => 'status' )->pack( -side => 'left' );
	my $w_status = $frame3->ROText( -width => 40, -height => 1 );
	$w_status->pack( -side => 'left', -expand => 1, -fill => 'both' );
	$w_balloon->attach( $w_status,
		-msg => 'display informations on internal state' );

	# for debuging
	$main->bind( '<Key-d>', [ \&dump_struct ] );

	# common gui subroutines with different coding
	##############################################
	# define insert_gui_orphans as a ref to anonymous sub
	*insert_gui_orphans = sub {
		my @orphans = @_;

		my $opt_arch = get_option('arch');

		$W_list->delete( 0, 'end' );

		## no critic (RequireInterpolationOfMetachars)
		my $qf =
		  ($opt_arch)
		  ? '%{NAME};%{VERSION}-%{RELEASE};%{ARCH};%{SIZE}\n'
		  : '%{NAME};%{SIZE}\n';
		my @orphans2 = rpm_query( $qf, @orphans );
		foreach (@orphans2) {
			chomp;
			$W_list->insert( 'end', [ split /;/ ] );
		}

		my $nb = scalar @orphans;
		$w_orphans->delete( '1.0', 'end' );
		$w_orphans->insert( '1.0', $nb );

		# reset
		$w_selected->delete( '1.0', 'end' );
		$w_selected->insert( '1.0', 0 );
	};

	*display_status = sub {
		my $text = shift @_;

		debug($text);

		$w_status->delete( '1.0', 'end' );
		$w_status->insert( '1.0', $text );

		# an update before the mainloop call cause a core dump
		# so we have to test if we are before of after
		$w_status->update() if $Inloop;
	};

	*display_total = sub {

		# display orphans total
		my $nb = scalar @Liste_pac;
		$w_total->delete( '1.0', 'end' );
		$w_total->insert( '1.0', $nb );
	};
	return;
}

#########################################################
1;
